#lang racket

;; Understanding the qualtrics data
;;
;; Rows in google.tsv ...
;;   1 = row titles
;;   2 = row Q detail
;;   3 = nonsense JS metadata???
;;   4 = real data begins here

(require
  gregor
  json
  (only-in racket/list first second third fourth partition)
  (only-in racket/string string-join)
  math/statistics
  gtp-util
  racket/runtime-path)

(module+ test (require rackunit))

(pretty-print-columns 102)

;; -----------------------------------------------------------------------------

(define-runtime-path PWD ".")
(define tmp-dir (build-path PWD "tmp"))
(define input-dir (build-path PWD ))
(define output-dir (build-path PWD "output"))

(define tmp/compiled (build-path tmp-dir "compiled"))

;(define orig.tsv (build-path input-dir "LTL_April 19, 2021_12.13.tsv"))
;(define google.tsv (build-path input-dir "google.tsv"))
(define edit.tsv (build-path input-dir "input_edit.tsv"))
(define qualtrics-raw.tsv edit.tsv)
(define prompt.rktd (build-path input-dir "prompt.rktd"))
(define forge-template.rkt (build-path input-dir "forge-template.rkt"))
(define equality-template.rkt (build-path input-dir "equality-template.rkt"))

(define tmp-forge.rkt (build-path tmp-dir "student-forge.rkt"))

(define (exe-exists? ps)
  (and (file-exists? ps)
       ;; TODO check executable permission
       #true))

(define (assert-exe ps)
  (if (exe-exists? ps)
    ps
    (raise-argument-error "cannot find executable" "attempt" ps)))

(define braco
  (assert-exe (build-path "/Applications" "Racket-7.9-BC" "bin" "raco")))

(define bracket
  (assert-exe (build-path "/Applications" "Racket-7.9-BC" "bin" "racket")))

;; ---

(define (check-tabs tsv)
  (define count*
    (with-input-from-file
      tsv
      (lambda ()
        (for/list ((ln (in-lines))
                   (i (in-naturals)))
                  (when (< i 3) (writeln ln))
          (count-tab ln)))))
  (define num-take 4)
  (cond
    ((apply = count*)
     (printf "all counts equal ~a~n" (car count*)))
    ((= (car count*) (seventh count*))
     (printf "first count = data count ~a~n" (car count*))
     (printf " ~s~n" (take count* num-take))
     (void))
    ((= (second count*) (seventh count*))
     (printf "second count = data count ~a~n" (car count*))
     (printf " ~s~n" (take count* num-take))
     (void))
    (else
     (printf "damn~n ~s~n" (take count* num-take))
     (void))))

(struct dataset [schema text row*] #:transparent)

(define (make-dataset schema text row*)
  (unless (and (pair? schema) (andmap string? schema) (all-distinct? schema))
    (raise-argument-error 'make-dataset "(listof string?) no-dups" 0 schema text row*))
  (unless (and (pair? text) (andmap string? text))
    (raise-argument-error 'make-dataset "(listof string?)" 1 schema text row*))
  (unless (and (pair? row*) (andmap pair? row*))
    (raise-argument-error 'make-dataset "(listof row?)" 2 schema text row*))
  (dataset schema text row*))

(define (analyze tsv)
  (define ds (tsv->dataset tsv))
  (define prompt* (file->value prompt.rktd))
;  (void
;    (explode-vertical
;      (build-path output-dir "per-student.txt")
;      prompt*
;      ds))
;  (void
;    (explode-horizontal
;      (build-path output-dir "per-question.txt")
;      prompt*
;      ds))
  (void
    (print-metadata
      (build-path output-dir "meta.txt")
      ds))
  (void))

(define (check-ltl-syntax tsv)
  (define ds (tsv->dataset tsv))
  (define prompt* (file->value prompt.rktd))
  (define s* (dataset-schema ds))
  (define rr* (dataset-row* ds))
  (with-output-to-file
    (build-path output-dir "bad-forge-answer.txt")
    #:exists 'replace
    (lambda ()
      (define row-ref (make-row-ref s*))
      (for ((ss (in-list s*))
            #:when (question-needs-validation? ss))
        (define bad-student**
          (for*/list ((rr (in-list rr*))
                      (fm (in-value (row-ref rr ss)))
                      #:unless (valid-forge-formula? fm))
            (list (row->email rr row-ref) fm)))
        (unless (null? bad-student**)
          (displayln ss)
          (prompt-print prompt* ss)
          (newline)
          (pretty-print bad-student**)
          (newline)
          (newline))))))

(define (check-english->ltl tsv)
  (define ds (tsv->dataset tsv))
  (define prompt* (file->value prompt.rktd))
  (define s* (dataset-schema ds))
  (define rr* (dataset-row* ds))
  (define num-students (length rr*))
  (with-output-to-file
    (build-path output-dir "E-L-bad2.html")
    #:exists 'replace
    (lambda ()
      (define row-ref (make-row-ref s*))
      (print-html-head)
      (for ((ss (in-list s*))
            #:when (and
                     (english->ltl-question? ss)
                     (question-needs-validation? ss)))
        (define staff-ans* (prompt-answer* (prompt-ref prompt* ss)))
        (define-values [good-student** bad-student**]
          (for*/lists (_1 _2 #:result (values (filter values _1) (filter values _2)))
                      ((rr (in-list rr*))
                       (fm (in-value (row-ref rr ss))))
            (define res (list (formula-format fm) (row->email rr row-ref) (row-ref rr (q+ ss 1)) (row-ref rr (q+ ss 2))))
            (if (correct-ltl? fm (car staff-ans*))
              (values res #f)
              (values #f res))))
        (question-print-html ss)
        (prompt-print-html prompt* ss)
        (let* ((num-bad (length bad-student**))
               (num-good (- num-students num-bad)))
          (print-count-html num-bad num-good num-students))
        (html-newline)
        (void
          (printf "<details open><summary>~a</summary><ul>~n" "correct answer")
          (print-student-info-html* good-student**)
          (printf "</ul></details>~n"))
        (let* ((grouped*
                (let loop ((bad-student** bad-student**)
                           (m* (append extra-miscon* (hash-ref ltl->misconception# ss '()))))
                  (cond
                    [(null? bad-student**)
                     '()]
                    [(null? m*)
                     (list (list "OTHER" bad-student**))]
                    [else
                      (define curr-m (car m*))
                      (define-values [here there]
                        (partition (has-misconception? curr-m staff-ans*) bad-student**))
                      (cons (list curr-m here) (loop there (cdr m*)))]))))
          (for ((g (in-list grouped*)))
            (printf "<details open><summary>~a students : ~a</summary><ul>~n" (length (cadr g)) (car g))
            (cond
              [(equal? "OTHER" (car g))
               (define g** (len-desc (map sort-answer2* (group-by car (cadr g) run-equality))))
               (if (or (null? g**) (null? (cdr g**)))
                 (print-student-info-html* (cadr g))
                 (for ((g* (in-list g**)))
                   (print-student-info-html* g*)
                   (printf "</ul><ul>~n")
                   (void)))]
              [else
               (print-student-info-html* (cadr g))])
            (printf "</ul></details>~n")
            (void)))
        (html-newline)
        (newline))
      (print-html-tail))))

(define (check-english->ltl-avg-confidence tsv)
  (define ds (tsv->dataset tsv))
  (define prompt* (file->value prompt.rktd))
  (define s* (dataset-schema ds))
  (define rr* (dataset-row* ds))
  (define num-students (length rr*))
  (with-output-to-file
    (build-path output-dir "E-L-tmp.txt")
    #:exists 'replace
    (lambda ()
      (define row-ref (make-row-ref s*))
      (print-html-head)
      (for ((ss (in-list s*))
            #:when (and
                     (english->ltl-question? ss)
                     (question-needs-validation? ss)))
        (printf
          "~a ~a~n"
          ss
          (mean
          (for/list ((rr (in-list rr*)))
            (string->number (car (string-split (row-ref rr (q+ ss 1))))))))
        (void)))))

(define (print-english->ltl tsv)
  (define ds (tsv->dataset tsv))
  (define prompt* (file->value prompt.rktd))
  (define s* (dataset-schema ds))
  (define rr* (dataset-row* ds))
  (define row-ref (make-row-ref s*))
  (for ((ss (in-list s*))
        #:when (and
                 (english->ltl-question? ss)
                 (question-needs-validation? ss)))
    (define qstr (first (prompt-ref prompt* ss)))
    (define staff-ans* (prompt-answer* (prompt-ref prompt* ss)))
    (for* ((rr (in-list rr*))
           (fm (in-value (row-ref rr ss))))
      (displayln
        (tab-join (list "2021"
                        (formula-format fm)
                        (email->id (row->email rr row-ref))
                        (format-confidence (row-ref rr (q+ ss 1)))
                        qstr
                        (car staff-ans*)
                        "staff"
                        (row-ref rr (q+ ss 2))))))))

(define (print-ltl->english tsv)
  (define ds (tsv->dataset tsv))
  (define prompt* (file->value prompt.rktd))
  (define s* (dataset-schema ds))
  (define rr* (dataset-row* ds))
  (define row-ref (make-row-ref s*))
  (for ((ss (in-list s*))
        #:when (ltl->english-question? ss))
    (define qstr (first (prompt-ref prompt* ss)))
    (define staff-ans
      (let ((pp (prompt-answer* (prompt-ref prompt* ss))))
        (if (pair? pp) (car pp) pp)))
    (for ((rr (in-list rr*)))
      (define fm (formula-format (row-ref rr ss)))
      (displayln
        (tab-join
          (list "2021"
                fm
                (email->id (row->email rr row-ref))
                (format-confidence (row-ref rr (q+ ss 1)))
                qstr
                staff-ans
                "staff"
                (row-ref rr (q+ ss 2))))))))

(define (print-traces tsv)
  (define ds (tsv->dataset tsv))
  (define prompt* (file->value prompt.rktd))
  (define s* (dataset-schema ds))
  (define rr* (dataset-row* ds))
  (define row-ref (make-row-ref s*))
  (for ((ss (in-list s*))
        #:when (and
                 (traces-part-question? ss)
                 (traces-boolean-question? ss)))
    (define pp (prompt-ref prompt* ss))
    (define qstr (first pp))
    (define tstr (second pp))
    (define staff-ans (boolean-format (prompt-answer* (prompt-ref prompt* ss))))
    (for ((rr (in-list rr*)))
      (define fm (formula-format (row-ref rr ss)))
      (define why (row-ref rr (q+ ss 1)))
      (displayln
        (tab-join
          (list fm
                why
                (email->id (row->email rr row-ref))
                qstr
                tstr
                staff-ans))))))

(define (tab-join str*)
  (string-join str* "\t"))

(define (print-count-html num-bad num-good num-students)
  (printf "<p>~a / ~a correct (~a%)</p>~n" num-good num-students (inexact->exact (* 100 (rnd (/ num-good num-students))))))

(define (has-misconception? curr-m staff-ans*)
  (cond
    [(string=? curr-m "syntax error")
     (lambda (st-info)
       (not (valid-forge-formula? (car st-info))))]
    [(string=? curr-m "inexpressible")
     (lambda (st-info)
       (define st-formula (car st-info))
       (string-contains? (string-downcase st-formula) "inexpressible"))]
    [(string=? curr-m "past-time")
     (lambda (st-info)
       (define st-formula (car st-info))
       (define fml (string-downcase st-formula))
       (for/or ((str (in-list '("before" "once" "historically"))))
         (string-contains? fml str)))]
    [(string=? curr-m "precedence")
     (define staff* (map tokenize staff-ans*))
     (lambda (st-info)
       (define st-formula (tokenize (car st-info)))
       (for/or ((staff (in-list staff*)))
         (equal? staff st-formula)))]
    [else
     (lambda (st-info)
       (define st-formula (car st-info))
       (run-equality st-formula curr-m))]))

(define (tokenize str)
  (let* ((str (string-replace* str
                               '(("=>" " implies ")
                                 ("{" " ")
                                 ("}" " ")
                                 ("[" " ")
                                 ("]" " ")
                                 ("(" " ")
                                 (")" " ")))))
    (string-split str)))

(module+ test
  (test-case "tokenize"
    (check-equal?
      (tokenize "{ Red not in Panel.lit } until { Red in Panel.lit and after { always { Red not in Panel.lit } } }")
      (string-split "Red not in Panel.lit until Red in Panel.lit and after always Red not in Panel.lit"))
    (check-equal?
      (tokenize "{Red in Panel.lit} until always { not Red in Panel.lit }")
      '("Red" "in" "Panel.lit" "until" "always" "not" "Red" "in" "Panel.lit"))
    (check-equal?
      (tokenize "{always Red in Panel.lit} until {always {not Red in Panel.lit}}")
      (string-split "always Red in Panel.lit until always not Red in Panel.lit"))))

(define (check-traces tsv)
  (define ds (tsv->dataset tsv))
  (define prompt* (file->value prompt.rktd))
  (define s* (dataset-schema ds))
  (define rr* (dataset-row* ds))
  (define num-students (length rr*))
  (with-output-to-file
    (build-path output-dir "traces-bad-answer.html")
    #:exists 'replace
    (lambda ()
      (print-html-head)
      (define row-ref (make-row-ref s*))
      (for ((ss (in-list s*))
            #:when (and
                     (traces-part-question? ss)
                     (traces-boolean-question? ss)))
        (define staff-ans (car (prompt-answer* (prompt-ref prompt* ss))))
        (define-values [good-student** bad-student**]
          (for*/lists (_1 _2 #:result (values (filter values _1) (filter values _2)))
                      ((rr (in-list rr*))
                       (fm (in-value (row-ref rr ss))))
            (define res (list (formula-format fm) (row->email rr row-ref) (row-ref rr (q+ ss 1))))
            (if (correct-traces? fm staff-ans)
              (values res #f)
              (values #f res))))
        (question-print-html ss)
        (prompt-print-html prompt* ss)
        (let* ((num-bad (length bad-student**))
               (num-good (- num-students num-bad)))
          (print-count-html num-bad num-good num-students))
        #;(for ((title (in-list '("correct, good answer" "correct, interesting answer" "incorrect, XXX" "incorrect, XXX"))))
          (printf "~n<h5>~a</h5>~n<details><ul>~n~n</ul></details>~n" title))
        (html-newline)
        (print-student-info-html* good-student**)
        (html-newline)
        (print-student-info-html* bad-student**)
        (html-newline)
        (newline))
    (print-html-tail))))

(define (format-confidence str)
  (car (string-split str)))

(define (formula-format str)
  (string-join (string-split str)))

(define (boolean-format b)
  (if b "Yes" "No"))

(define (question-print-html ss)
  (printf "<h4>~a</h4>~n" ss))

(define (html-newline)
  (printf "~n<br/>~n"))

(define (print-student-info-html* st-info*)
  (for ((st (in-list st-info*)))
    (printf "<li>~a &nbsp;&nbsp; ~a &nbsp;&nbsp; ~a</li>~n"
            (car st) (format-email-html (cadr st))
            (string-join (cddr st) " &nbsp;&nbsp; "))))

(define (print-student-table-html* st-info*)
  (for ((st (in-list st-info*)))
    (printf "<tr><td>~a &nbsp;&nbsp; </td><td> ~a &nbsp;&nbsp; </td><td>  ~a &nbsp;&nbsp; </td><td><div class=\"hidden\"><br/> ~a <br/></div></td></tr>~n"
            (format-email-html (cadr st))
            (third st)
            (first st)
            (fourth st))))

(define past-time-span-str #<<EOS
<span class="rubric pt">past</span>
EOS
)

(define (format-email-html pre-str)
  (define str (string-trim pre-str))
  (format "<a href=\"mailto:~a\">~a</a>" str (email->id str)))

(define (email->id str)
  (substring+ (string-trim str) 0 4))

(define (substring+ str lo hi)
  (if (<= (string-length str) hi)
    str
    (substring str lo hi)))

(define (print-student-info* st-info*)
  (for ((st (in-list st-info*)))
    (printf "~a~n ~a~n" (car st) (cadr st))))

(define (print-html-head)
  (displayln #<<EOS
<html>
  <head>
    <style>
    .rubric { text-color: #000000; padding: 2px }
    .wu { background-color: #ff8833 }
    .nu { background-color: #ff9944 }
    .gu { background-color: #ee9955 }
    .ug { background-color: #ff9966 }
    .au { background-color: #ff9977 }

    .pp { background-color: #ff99cc }
    .ig { background-color: #ffcccc }
    .if { background-color: #ffcccc }
    .ll { background-color: #ffff77 }
    .nl { background-color: #eeee99 }

    .xc { background-color: #00ff00 }
    .sx { background-color: #00ff00 }
    .nx { background-color: #00ff00 }

    .fg { background-color: #aaaaff }
    .gf { background-color: #aaaaff }
    .ui { background-color: #aaaaff }
    .iu { background-color: #aaaaff }
    .ai { background-color: #aaaaff }

    .ji { background-color: #ff8888 }
    .pt { background-color: #666666 }

    .hidden { display: none }
    table { margin-left: 20px; margin-bottom: 16px }
    </style>
  </head>
<body>

<h4>Rubric</h4>
<table>
  <tr><td><span class="rubric wu">weak U</span></td><td>weak until</td></tr>
  <tr><td><span class="rubric nu">U neg</span></td><td>U requires negation on rhs</td></tr>
  <tr><td><span class="rubric ug">G U</span></td><td>U extra G on lhs</td></tr>
  <tr><td><span class="rubric gu">U G</span></td><td>U implicit G on rhs</td></tr>
  <tr><td><span class="rubric au">G(U)</span></td><td>always { a until a }</td></tr>

  <tr><td><span class="rubric pp">( )</span></td><td>paren error, bad association</td></tr>
  <tr><td><span class="rubric ig">imp G</span></td><td>implicit G</td></tr>
  <tr><td><span class="rubric if">imp F</span></td><td>implicit F</td></tr>
  <tr><td><span class="rubric ll">lasso</span></td><td>lasso mixup</td></tr>
  <tr><td><span class="rubric nl">no lasso</span></td><td>no lasso, 5 states is entire trace</td></tr>

  <tr><td><span class="rubric xc">num X</span></td><td>X mis-count</td></tr>
  <tr><td><span class="rubric sx">span X</span></td><td>X applies for current and next state</td></tr>
  <tr><td><span class="rubric nx">neg X</span></td><td>X negates in current state</td></tr>

  <tr><td><span class="rubric fg">FG</span></td><td>swap F to G</td></tr>
  <tr><td><span class="rubric gf">GF</span></td><td>swap G to F</td></tr>
  <tr><td><span class="rubric ui">U=></span></td><td>swap U to implication</td></tr>
  <tr><td><span class="rubric iu">=>U</span></td><td>swap implication to U</td></tr>
  <tr><td><span class="rubric ai">/\=></span></td><td>swap and to implication</td></tr>

  <tr><td><span class="rubric ji">=> jump</span></td><td>implication jumps to a state where lhs. sat.</td></tr>
  <tr><td><span class="rubric pt">past</span></td><td>past-time/illegal op</td></tr>
</table>
<br/>
<br/>
<br/>

EOS
))

(define (print-html-tail)
  (printf "~n</body></html>~n~n"))

(define (check-ltl->english tsv)
  (define ds (tsv->dataset tsv))
  (define prompt* (file->value prompt.rktd))
  (define s* (dataset-schema ds))
  (define rr* (dataset-row* ds))
  (define num-students (length rr*))
  (with-output-to-file
    (build-path output-dir "L-E-bad-answer.html")
    #:exists 'replace
    (lambda ()
      (define row-ref (make-row-ref s*))
      (print-html-head)
      (for ((ss (in-list s*))
            #:when (ltl->english-question? ss))
        (define staff-ans (car (prompt-answer* (prompt-ref prompt* ss))))
        (define st-info*
          (for/list ((rr (in-list rr*)))
            (define fm (formula-format (row-ref rr ss)))
            (list fm (row->email rr row-ref) (format-confidence (row-ref rr (q+ ss 1))) (row-ref rr (q+ ss 2)))))
        (question-print-html ss)
        (prompt-print-html prompt* ss)
        (html-newline)
        (printf "<p>~a average confidence</p>~n"
                (~r (mean (map (compose1 string->number third) st-info*)) #:precision '(= 2)))
        (html-newline)
        (void
          (printf "<details open><summary>~a</summary><table>~n" "correct answer")
          (printf "</table></details>~n"))
        (html-newline)
        (newline)
        (printf "<table>~n")
        (print-student-table-html* st-info*)
        (printf "</table>~n")
        (html-newline)
        (newline))
      (print-html-tail))))

(define (explode-vertical ps prompt* ds)
  (unless (file-exists? ps)
    (with-output-to-file
      ps
      (lambda ()
        (define s* (dataset-schema ds))
        (define t* (dataset-text ds))
        (define rr* (dataset-row* ds))
        (define row-ref (make-row-ref s*))
        (for ((rr (in-list rr*)))
          (for ((ss (in-list s*))
                (tt (in-list t*))
                #:when (basic-question? ss))
            (displayln ss)
            (prompt-print prompt* ss)
            (displayln tt)
            (newline)
            (pretty-print (row-ref rr ss))
            (newline)
            (newline)
            (void)))))))

(define (explode-horizontal ps prompt* ds)
  (unless (file-exists? ps)
    (with-output-to-file
      ps
      (lambda ()
        (define s* (dataset-schema ds))
        (define t* (dataset-text ds))
        (define rr* (dataset-row* ds))
        (define row-ref (make-row-ref s*))
        (for ((ss (in-list s*))
              (tt (in-list t*))
              #:when (basic-question? ss))
          (displayln ss)
          (prompt-print prompt* ss)
          (displayln tt)
          (newline)
          (for ((rr (in-list rr*)))
            (pretty-print (row-ref rr ss))
            (void))
          (newline)
          (newline)
          (void))))))

(define (print-metadata ps ds)
  (unless (file-exists? ps)
    (with-output-to-file
      ps
      (lambda ()
        (define s* (dataset-schema ds))
        (define t* (dataset-text ds))
        (define rr* (dataset-row* ds))
        (define row-ref (make-row-ref s*))
        (define (row-ref* str) (map (lambda (rr) (row-ref rr str)) rr*))
        (define submit-time-question* (filter (lambda (ss) (string-contains? ss "Submit")) s*))
        ;; key stats
        (void
          (print-date-stats "Start - End" (row-ref* "StartDate") (row-ref* "EndDate"))
          (newline)
          (print-seconds-stats "Duration" (map string->seconds (row-ref* "Duration (in seconds)")))
          (newline)
          (let ((t* (for/list ((rr (in-list rr*)))
                      (apply + (map string->seconds (for/list ((ss (in-list submit-time-question*))) (row-ref rr ss)))))))
            (print-seconds-stats "D2" t*))
          (newline)
          (for ((txt (in-list submit-time-question*)))
            (print-seconds-stats txt (map string->seconds (row-ref* txt)))
            (newline)))
          ;; everything else
          (printf "~n~n@@@@~n~n")
          (for ((ss (in-list s*))
                (tt (in-list t*))
                #:unless (basic-question? ss))
            (pretty-print ss)
            (pretty-print tt)
            (newline)
            (for ((rr (in-list rr*)))
              (pretty-print (row-ref rr ss))
              (void))
            (newline)
            (newline)
            (void))))))

(define (cluster-ltl* txt)
  (define qa* (parse-qa-file txt))
  (define qa**
    (for/list ((qa (in-list qa*)))
      (define q (car qa))
      (define a* (cdr qa))
      (define-values [good-a* bad-a*] (partition (compose1 valid-forge-formula? ltl-ans->formula) a*))
      (cons q
            (map sort-answer*
                 (cons bad-a*
                       (group-by ltl-ans->formula good-a* run-equality))))))
  (with-output-to-file
    (build-path output-dir "E-L-cluster.txt")
    #:exists 'replace
    (lambda ()
      (for ((qa (in-list qa**)))
        (displayln (car qa))
        (for ((a (in-list (cdr qa))))
          (for-each displayln a)
          (newline))
        (newline))))
  (void))

(define (autograde-2020 txt)
  (define qa**
    (with-input-from-file
      txt
      (lambda ()
        (let loop ((acc '()))
          (define v (read-line))
          (cond
            [(eof-object? v)
             (reverse acc)]
            [(empty-string? v)
             (loop acc)]
            [(string-prefix? v "-----")
             (loop (cons (reverse (car acc)) (cdr acc)))]
            [(string-prefix? v "Prompt")
             (define _rationale (read-line))
             (define correct (substring (read-line) (string-length "Correct:")))
             (define _ans (read-line))
             (loop (cons (list (list v correct)) acc))]
            [(parse-json v)
             => (lambda (jj)
             (loop (cons (cons jj (car acc)) (cdr acc))))]
            [else
              (printf "BAD~n ~s~n" v)
              (raise-arguments-error 'autograde "parse error" "line" v)])))))
  (with-output-to-file "brotmptmp" #:exists 'replace (lambda () (pretty-write qa**)))
  (define graded-qa**
    (for/list ((qa* (in-list qa**)))
      (define title (car qa*))
      (define the-ans (second title))
      (define all-a* (map (lambda (a) (cons a (jsexpr-2020->electrum a))) (cdr qa*)))
      (define-values [correct-a* incorrect-a*]
        (partition (lambda (x)
                     (and (cdr x)
                          (run-equality (cdr x) the-ans)))
                   all-a*))
      (list title
            correct-a*
            incorrect-a*)))
  (with-output-to-file
    (build-path output-dir "2020-autograde.txt")
    #:exists 'replace
    (lambda ()
      (for ((qa* (in-list graded-qa**)))
        (define title (first qa*))
        (define good-a* (second qa*))
        (define bad-a* (third qa*))
        (define num-good (length good-a*))
        (define num-bad (length bad-a*))
        (printf "~a~n~a~n ~a / ~a correct (~a%)~n~n"
                (first title) (second title)
                num-good (+ num-good num-bad)
                (rnd (* 100 (/ num-good (+ num-good num-bad)))))
        (for-each print-2020-ans good-a*)
        (newline)
        (for-each print-2020-ans bad-a*)
        (newline)
        (newline)))))

(define (print-2020-ans a)
  (printf "~a\t~a~n" (jsexpr->string (car a)) (cdr a)))

(define (jsexpr-2020->electrum jj)
  (define ans (hash-ref jj 'answer #f))
  (and ans (not (eq? 'null ans)) (token-subst-2020* ans)))

(define (token-subst-2020* str)
  (string-replace*
    str
    '(("G" " always ")
      ("F" " eventually ")
      ("U" " until ")
      ("X" " after ")
      ("!" " not ")
      ("~" " not ")
      ("^" " and ")
      ("<-->" "<=>")
      ("->" "=>")
      ("/\\" " and ")
      ("\\/" " or ")
      ("x1" " Red in Panel.lit ")
      ("x2" " Blue in Panel.lit "))))

(define (string-replace* str kv*)
  (for/fold ((acc str))
            ((kv (in-list kv*)))
    (string-replace acc (car kv) (cadr kv))))

;; ---

(define (len-desc x**)
  (sort x** > #:key length))

(define (sort-answer* str* #:key [key values])
  (sort str* string<? #:key key))

(define (sort-answer2* str*)
  (sort-answer* str* #:key car))

(define (ltl-ans->formula str)
  (cadr (tab-split str)))

(define (parse-qa-file txt)
  (with-input-from-file
    txt
    (lambda ()
      (for/fold ((acc '()) #:result (reverse acc))
                ((ln (in-lines)))
        (cond
          [(question-string? ln)
           (cons (cons ln '()) acc)]
          [(empty-string? ln)
           acc]
          [else
           (define fst (car acc))
           (cons (cons (car fst) (cons ln (cdr fst)))
                 (cdr acc))])))))

(define (empty-string? str)
  (zero? (string-length str)))

(define string-empty? empty-string?)

(define (prompt-ref p* key)
  (or
    (for/first ((q-info (in-list p*))
                #:when (equal? (q-info->q-num q-info) key))
      (q-info->rest q-info))
    '()))

(define (prompt-answer* r*)
  (if (null? r*) (raise-argument-error 'prompt-answer "pair?" r*) (last r*)))

(define (prompt-print p* q-key)
  (define rr* (prompt-ref p* q-key))
  (cond
    [(null? rr*)
     (void)]
    [(or
       (ltl->english-part-question? q-key)
       (english->ltl-part-question? q-key))
     (printf "~a~nA. ~a~n" (first rr*) (second rr*))]
    [(traces-part-question? q-key)
     (printf "formula: ~a~n trace: ~a~n answer ~a~n" (first rr*) (second rr*) (third rr*))]
    [else
      (for-each displayln rr*)]))

(define (prompt-print-html p* q-key)
  (define rr* (prompt-ref p* q-key))
  (cond
    [(null? rr*)
     (void)]
    [(or
       (ltl->english-part-question? q-key)
       (english->ltl-part-question? q-key))
     (printf "<table>~n<tr><td>Q</td><td>~a</td></tr>~n<tr><td>A</td><td>~a</td></tr></table>~n" (first rr*) (car (second rr*)))]
    [(traces-part-question? q-key)
     (printf "<table>~n<tr><td>formula:</td><td>~a</td></tr>~n<tr><td>trace:</td><td>~a</td></tr>~n<tr><td>answer</td><td>~a</td></tr></table>~n" (first rr*) (second rr*) (car (third rr*)))]
    [else
      (for-each displayln rr*)]))

(define q-info->q-num car)
(define q-info->rest cdr)

(define (string->seconds str)
  (exact-round (string->number str)))

(define (bytes-count-tab str)
  (for/sum ((c (in-bytes str)))
    (if (eq? c #\tab) 1 0)))

(define (count-tab str)
  (for/sum ((c (in-string str)))
    (if (eq? c #\tab) 1 0)))

(define (parse-tsv tsv)
  (with-input-from-file
    tsv
    (lambda ()
      (for/list ((ln (in-lines)))
        (tab-split ln)))))

(define (tab-split str)
  (string-split str "\t"))

(define (all-distinct? x*)
  (= (length x*)
     (length (remove-duplicates x*))))

(define ((make-row-ref schema) rr kk)
  (or
    (for/first ((r (in-list rr))
                (s (in-list schema))
                #:when (equal? s kk))
      r)
    (raise-arguments-error 'row-ref "key not found in schema" "key" kk "schema" (~e schema) "row" (~e rr))))

(define row->email
  (case-lambda
    ((rr)
     (list-ref rr 18))
    ((rr ref)
     (ref rr "Q1.7"))))

(module+ test
  (test-case "row->email"
    (check-equal? (row->email '("2021-04-09 18:55:52" "2021-04-09 19:55:10" "IP Address" "100" "3557" "True" "2021-04-09 19:55:10" "R_1rwE5OrOOuIeigR" "anonymous" "EN" "Chrome" "89.0.4389.114" "Macintosh" "1440x900" "5.32" "366.142" "370.408" "12" "917u0cvrthn6oh5a@gmail.com" "If the red light is on then three states from now it must be on " "5 (Highly Confident)" "A possible misconception is the light stays on for all intermediate states" "Two states from now there will be some future state such that the state after it will have the red light on.   To put in more mathematical terms (sorry) but in pseudo English : If the current state is i then there is a state i+2 and a state x such that x is greater than or equal  to i+2 such that in x+1 the red light is on" "3 (Somewhat Confident)" "There could be a lot of confusion about the nesting of the operators and what it means such as thinking that the state after has the red light on or that red light is on 3 states from now." "If the red light will be on now or in some future state that means the blue light must always be on." "4"))
                  "917u0cvrthn6oh5a@gmail.com")
    (check-equal? (row->email '("A" "B" "C")
                              (make-row-ref '("QQ" "Q1.7" "blah")))
                  "B")))

(define filter-ignore-row
  (let* ((black-str "{\"ImportId\":\"startDate\",\"timeZone\":\"America/Denver\"}")
         (black-email* (list "benjaminlgreenman@gmail.com" "tbnelson@gmail.com")))
    (lambda (ds)
      (define row-ref (make-row-ref (dataset-schema ds)))
      (make-dataset
        (dataset-schema ds)
        (dataset-text ds)
        (for/list ((rr (in-list (dataset-row* ds)))
                   #:unless (or (equal? (first rr) black-str)
                                (member (row-ref rr "Q1.7") black-email*)
                                (equal? (row-ref rr "Finished") "False")))
          rr)))))

(define (basic-question? str)
  (regexp-match? #rx"^Q[0-9]+\\.[0-9]+(_[0-9]+)?$" str))

(module+ test
  (test-case "basic-question?"
    (check-true (basic-question? "Q2.7"))
    (check-true (basic-question? "Q2.23"))
    (check-true (basic-question? "Q6.22_2"))

    (check-false (basic-question? "StartDate"))
    (check-false (basic-question? "Q1.1_Version"))
    (check-false (basic-question? "Q4.22_Page Submit"))
    (check-false (basic-question? "Q2.22_First Click"))))

(define (print-date-stats title pre-start* pre-end*)
  (define start* (map qualtrics->datetime pre-start*))
  (define end* (map qualtrics->datetime pre-end*))
  (define diff* (map seconds-between start* end*))
  (print-seconds-stats title diff*))

(define (qualtrics->datetime str)
  (iso8601->datetime (string-replace str " " "T")))

(define (print-seconds-stats title pre-t*)
  (define t* (map sec->min pre-t*))
  (printf "~a (min)~n min ~a~n max ~a~n median ~a~n mean ~a~n stddev ~a~n"
          title
          (apply min t*)
          (apply max t*)
          (median < t*)
          (rnd (mean t*))
          (rnd (stddev t*))))

(define (sec->min t)
  (quotient t 60))

(define (equality-template/subst kv*)
  (template/subst equality-template.rkt kv*))

(define (forge-template/myFormula str)
  (template/subst forge-template.rkt (list (list "TODO" str))))

(define (template/subst fn kv*)
  (with-output-to-file
    tmp-forge.rkt
    #:exists 'replace
    (lambda ()
      (with-input-from-file
        fn
        (lambda ()
          (for ((ln (in-lines)))
            (let* ((str (string-trim ln))
                   (ln
                     (or (for/first ((kv (in-list kv*))
                                     #:when (equal? str (car kv)))
                           (cadr kv))
                         ln)))
              (displayln ln)))))))
  tmp-forge.rkt)

#;(module+ test
  #;(forge-template/myFormula "WE PA")
  (forge-template/subst (list (list "TODO" "WEA")))
  (error 'stop))

(define (correct-ltl? student staff)
  (and
    (valid-forge-formula? student)
    (run-equality student staff)))

(define (run-equality student staff)
  (run-ok? (equality-template/subst
             (list (list "STUDENT" student)
                   (list "STAFF" staff)))))

(define (correct-traces? student staff)
  (eq? (yes-no->boolean student) staff))

(define (yes-no->boolean str)
  (cond
    [(equal? str "Yes") #true]
    [(equal? str "No") #false]
    [else (raise-argument-error 'yes-no->boolean "(or/c \"Yes\" \"No\")" str)]))

(define (valid-forge-formula? str)
  (and
    (not (string-contains? str "#lang"))
    (compile-ok? (forge-template/myFormula str))))

(define (compile-ok? file.rkt)
  (rm-rf tmp/compiled)
  (system* braco "make" file.rkt))

(define (run-ok? file.rkt)
  (with-output-to-file
    "/dev/null"
    #:exists 'append
    (lambda ()
      (rm-rf tmp/compiled)
      (and
        (system* braco "make" file.rkt)
        (system* bracket file.rkt)))))

(define (rm-rf ps)
  (delete-directory/files ps #:must-exist? #false))

(define (tsv->dataset tsv)
  (let* ((row* (parse-tsv tsv))
         (ds (make-dataset (first row*) (second row*) (drop row* 3)))
         (ds (filter-ignore-row ds)))
    ds))

(define (question-string? str)
  (string-prefix? str "Q"))

(define (ltl->english-part-question? str)
  (string-prefix? str "Q2."))

(define (english->ltl-part-question? str)
  (string-prefix? str "Q4."))

(define (traces-part-question? str)
  (string-prefix? str "Q6."))

(define question-needs-validation*
  '("Q2.23" "Q4.3" "Q4.5" "Q4.7" "Q4.9" "Q4.11" "Q4.13" "Q4.15" "Q4.17" "Q4.19"
    "Q4.21" "Q6.22_1" "Q6.22_2"))

(define (question-needs-validation? str)
  (and (member str question-needs-validation*) #true))

(define question-traces-yes-no*
  '("Q6.3" "Q6.5" "Q6.7" "Q6.9" "Q6.11" "Q6.13" "Q6.15" "Q6.17" "Q6.19"))

(define (traces-boolean-question? ss)
  (and (member ss question-traces-yes-no*) #true))

(define english->ltl*
  '("Q4.3" "Q4.7" "Q4.11" "Q4.15" "Q4.19"))

(define (english->ltl-question? ss)
  (and (member ss english->ltl*) #true))

(define ltl->english*
  '("Q2.3" "Q2.7" "Q2.11" "Q2.15" "Q2.19"))

(define (ltl->english-question? ss)
  (and (member ss ltl->english*) #true))

(define extra-miscon*
  '("inexpressible" "syntax error" "precedence" "past-time"))

(define ltl->misconception#
  (hash
    "Q4.3"  (list "always { Red in Panel.lit implies not { after { Red in Panel.lit } } and after { after { Red in Panel.lit } } }"
                  "Red in Panel.lit implies after { Red not in Panel.lit } and after { after { Red in Panel.lit } }"
                  "always { Red in Panel.lit implies after { Red in Panel.lit }} and always { Red not in Panel.lit implies after { Red in Panel.lit }}"
                  "always { Red in Panel.lit until { after { Red not in Panel.lit } and after { after { Red in Panel.lit }}}}")
    "Q4.7"  (list "always { Red in Panel.lit implies always { after { Red not in Panel.lit }}}"
                  ;; #f ;; expected inexpressible
                  "{ Red not in Panel.lit } until { Red in Panel.lit and after { Red not in Panel.lit }}"
                  "eventually { Red in Panel.lit }"
                  "{ Red not in Panel.lit } until { Red in Panel.lit and after { eventually { Red not in Panel.lit }}}")
    "Q4.11" (list "not { Red in Panel.lit and after { Red in Panel.lit } and after { after { Red in Panel.lit }}}"
                  "always { not { Red in Panel.lit implies after { Red in Panel.lit } implies after { after { Red in Panel.lit }}}}"
                  "always { not { after { after { after { Red in Panel.lit }}}}}")
    "Q4.15" (list "always { Red in Panel.lit implies after { eventually { Blue in Panel.lit }}}"
                  "Red in Panel.lit implies after { eventually { Blue in Panel.lit }}"
                  "Red in Panel.lit implies { eventually { Blue in Panel.lit }}")
    "Q4.19" (list "Red in Panel.lit until { Red not in Panel.lit }"
                  "Red in Panel.lit until {}"
                  "Red in Panel.lit and { Red in Panel.lit implies always { Red not in Panel.lit }}"
                  "Red in Panel.lit implies eventually { always { Red not in Panel.lit }}"
                  "Red in Panel.lit and { Red in Panel.lit until { always { Red not in Panel.lit }}}"
                  "{ Red in Panel.lit until Red not in Panel.lit } and eventually { always { Red not in Panel.lit }}"
                  "Red in Panel.lit and { Red in Panel.lit until Red not in Panel.lit } and eventually { always { Red not in Panel.lit }}"
                  "eventually { always { Red not in Panel.lit }}"
                  ;; #f ;; expected inexpressible
                  )))

(define (q+ str n)
  (define-values [pre post]
    (let* ((str* (string-split str ".")))
      (values (car str*) (cadr str*))))
  (format "~a.~a" pre (+ (string->number post) n)))

(define historic-kw* '("once" "'" "historically" "before" "never"))

(define (historic-str? str)
  (for/or ((hh (in-list historic-kw*)))
    (string-contains? str hh)))

(define (parse-json str)
  (with-handlers ((exn:fail? (lambda (x) #f)))
    (string->jsexpr str)))

(define (quick-check-optional txt)
  (define qa*
    (with-input-from-file
      txt
      (lambda ()
        (let loop ((acc '()))
          (define v (read-line))
          (cond
            [(eof-object? v)
             (reverse (cons (reverse (car acc)) (cdr acc)))]
            [(string-prefix? v "Q")
             (let ((acc (if (null? acc) acc (cons (reverse (car acc)) (cdr acc)))))
               (void (read-line))
               (loop (cons (list v) acc)))]
            [(string-empty? v)
             (loop acc)]
            [else
              (loop (cons (cons (string->value v) (car acc)) (cdr acc)))])))))
  (let* ((all-a* (apply append (map cdr qa*))))
    (define-values [bad* good*]
      (partition (lambda (x) (string-empty? (string-trim x))) all-a*))
    (define num-good (length good*))
    (define num-all (length all-a*))
    (printf "overall = ~a / ~a (~a%) optional responses~n"
            num-good
            num-all
            (rnd (* 100 (/ num-good num-all)))))
  (for ((qa (in-list qa*)))
    (define all-a* (cdr qa))
    (define-values [bad* good*]
      (partition (lambda (x) (string-empty? (string-trim x))) all-a*))
    (define num-good (length good*))
    (define num-all (length all-a*))
    (printf " ~a : ~a / ~a (~a%)~n" (car qa) num-good num-all
            (rnd (* 100 (/ num-good num-all)))))
  (void))

;; -----------------------------------------------------------------------------

(module+ main

  #;(check-tabs qualtrics-raw.tsv)
  #;(analyze qualtrics-raw.tsv)
  #;(check-ltl-syntax qualtrics-raw.tsv)
  #;(check-english->ltl qualtrics-raw.tsv)
  #;(check-traces qualtrics-raw.tsv)
  #;(check-english->ltl-avg-confidence qualtrics-raw.tsv)
  #;(check-ltl->english qualtrics-raw.tsv)
  #;(cluster-ltl* (build-path input-dir "E-L-other.txt"))
  #;(autograde-2020 (build-path input-dir "2020-autograde.txt"))
  #;(quick-check-optional (build-path output-dir "per-optional.txt"))
  #;(print-english->ltl qualtrics-raw.tsv)
  #;(print-ltl->english qualtrics-raw.tsv)
  (print-traces qualtrics-raw.tsv)

  (void))

