#lang racket

;; Understanding the qualtrics data from Summer 2021 (round 3)

(require
  (only-in racket/list first second third fourth partition make-list)
  (only-in racket/string string-join non-empty-string?)
  math/statistics
  gtp-util
  racket/runtime-path)

(module+ test (require rackunit))

(pretty-print-columns 102)

;; -----------------------------------------------------------------------------

(define-runtime-path PWD ".")
(define input-dir PWD)

(define edit.tsv (build-path input-dir "input_raw.tsv"))
(define qualtrics-raw.tsv edit.tsv)
(define prompt.rktd (build-path input-dir "mid_prompt.rktd"))

(define (exe-exists? ps)
  (and (file-exists? ps)
       ;; TODO check executable permission
       #true))

;; ---

(struct dataset [schema text row*] #:transparent)

(define (make-dataset schema text row*)
  (unless (and (pair? schema) (andmap string? schema) (all-distinct? schema))
    (raise-argument-error 'make-dataset "(listof string?) no-dups" 0 schema text row*))
  (unless (and (pair? text) (andmap string? text))
    (raise-argument-error 'make-dataset "(listof string?)" 1 schema text row*))
  (unless (and (pair? row*) (andmap pair? row*))
    (raise-argument-error 'make-dataset "(listof row?)" 2 schema text row*))
  (dataset schema text (map (lambda (rr) (pad-to rr (length schema) "")) row*)))

(define (pad-to x* len default)
  (define ntail
    (- len (length x*)))
  (if (< 0 ntail)
    (append x* (make-list ntail default))
    x*))

(define Q_AREA "Q44")
(define Q_AREA_EXTRA "Q44_14_TEXT")

(define (print-ltl->english tsv)
  (define ds (tsv->dataset tsv))
  (define prompt* (file->value prompt.rktd))
  (define s* (dataset-schema ds))
  (define rr* (dataset-row* ds))
  (define row-ref (make-row-ref s*))
  (for ((ss (in-list s*))
        #:when (ltl->english-question? ss))
    (define qstr (first (prompt-ref prompt* ss)))
    (define staff-ans
      (let ((pp (prompt-answer* (prompt-ref prompt* ss))))
        (if (pair? pp) (car pp) pp)))
    (for ((rr (in-list rr*)))
      (define fm (formula-format (row-ref rr ss)))
      (displayln
        (tab-join
          (list "2021"
                fm
                (email->id (row->email rr row-ref))
                (format-area (row-ref rr Q_AREA) (row-ref rr Q_AREA_EXTRA))
                qstr
                staff-ans))))))

(define (print-english->ltl tsv)
  (define ds (tsv->dataset tsv))
  (define prompt* (file->value prompt.rktd))
  (define s* (dataset-schema ds))
  (define rr* (dataset-row* ds))
  (define row-ref (make-row-ref s*))
  (for ((ss (in-list s*))
        #:when (english->ltl-question? ss))
    (define qstr (first (prompt-ref prompt* ss)))
    (define staff-ans* (prompt-answer* (prompt-ref prompt* ss)))
    (for* ((rr (in-list rr*))
           (fm (in-value (row-ref rr ss))))
      (displayln
        (tab-join (list "2021"
                        (formula-format fm)
                        (email->id (row->email rr row-ref))
                        (format-area (row-ref rr Q_AREA) (row-ref rr Q_AREA_EXTRA))
                        qstr
                        (car staff-ans*)))))))

(define (print-traces tsv)
  (define ds (tsv->dataset tsv))
  (define prompt* (file->value prompt.rktd))
  (define s* (dataset-schema ds))
  (define rr* (dataset-row* ds))
  (define row-ref (make-row-ref s*))
  (for ((ss (in-list s*))
        #:when (traces-boolean-question? ss))
    (define pp (prompt-ref prompt* ss))
    (define qstr (first pp))
    (define fstr (second pp))
    (define staff-ans (prompt-answer* pp))
    (for* ((rr (in-list rr*))
           (fm (in-value (row-ref rr ss))))
      (displayln
        (tab-join (list (formula-format fm)
                        (email->id (row->email rr row-ref))
                        (format-area (row-ref rr Q_AREA) (row-ref rr Q_AREA_EXTRA))
                        qstr
                        fstr
                        (format-boolean staff-ans)))))))

(define (format-boolean b)
  (if b "Yes" "No"))

(define (tab-join str*)
  (string-join str* "\t"))

#;(define (check-traces tsv)
  (define ds (tsv->dataset tsv))
  (define prompt* (file->value prompt.rktd))
  (define s* (dataset-schema ds))
  (define rr* (dataset-row* ds))
  (define num-students (length rr*))
  (with-output-to-file
    (build-path output-dir "traces-bad-answer.html")
    #:exists 'replace
    (lambda ()
      (print-html-head)
      (define row-ref (make-row-ref s*))
      (for ((ss (in-list s*))
            #:when (and
                     (traces-part-question? ss)
                     (traces-boolean-question? ss)))
        (define staff-ans (car (prompt-answer* (prompt-ref prompt* ss))))
        (define-values [good-student** bad-student**]
          (for*/lists (_1 _2 #:result (values (filter values _1) (filter values _2)))
                      ((rr (in-list rr*))
                       (fm (in-value (row-ref rr ss))))
            (define res (list (formula-format fm) (row->email rr row-ref) (row-ref rr (q+ ss 1))))
            (if (correct-traces? fm staff-ans)
              (values res #f)
              (values #f res))))
        (question-print-html ss)
        (prompt-print-html prompt* ss)
        (let* ((num-bad (length bad-student**))
               (num-good (- num-students num-bad)))
          (print-count-html num-bad num-good num-students))
        #;(for ((title (in-list '("correct, good answer" "correct, interesting answer" "incorrect, XXX" "incorrect, XXX"))))
          (printf "~n<h5>~a</h5>~n<details><ul>~n~n</ul></details>~n" title))
        (html-newline)
        (print-student-info-html* good-student**)
        (html-newline)
        (print-student-info-html* bad-student**)
        (html-newline)
        (newline))
    (print-html-tail))))

(define (format-area str1 str2)
  (if (non-empty-string? str2)
    (string-append str1 "," str2)
    str1))

(define (formula-format str)
  (string-join (string-split str)))

(define (email->id str)
  (substring+ (string-trim str) 2 6))

(define (substring+ str lo hi)
  (if (<= (string-length str) hi)
    str
    (substring str lo hi)))

(define (print-student-info* st-info*)
  (for ((st (in-list st-info*)))
    (printf "~a~n ~a~n" (car st) (cadr st))))

(define (token-subst-2020* str)
  (string-replace*
    str
    '(("G" " always ")
      ("F" " eventually ")
      ("U" " until ")
      ("X" " after ")
      ("!" " not ")
      ("~" " not ")
      ("^" " and ")
      ("<-->" "<=>")
      ("->" "=>")
      ("/\\" " and ")
      ("\\/" " or ")
      ("x1" " Red in Panel.lit ")
      ("x2" " Blue in Panel.lit "))))

(define (string-replace* str kv*)
  (for/fold ((acc str))
            ((kv (in-list kv*)))
    (string-replace acc (car kv) (cadr kv))))

;; ---

(define (len-desc x**)
  (sort x** > #:key length))

(define (sort-answer* str* #:key [key values])
  (sort str* string<? #:key key))

(define (sort-answer2* str*)
  (sort-answer* str* #:key car))

(define (ltl-ans->formula str)
  (cadr (tab-split str)))

(define (parse-qa-file txt)
  (with-input-from-file
    txt
    (lambda ()
      (for/fold ((acc '()) #:result (reverse acc))
                ((ln (in-lines)))
        (cond
          [(question-string? ln)
           (cons (cons ln '()) acc)]
          [(empty-string? ln)
           acc]
          [else
           (define fst (car acc))
           (cons (cons (car fst) (cons ln (cdr fst)))
                 (cdr acc))])))))

(define (empty-string? str)
  (zero? (string-length str)))

(define string-empty? empty-string?)

(define (prompt-ref p* key)
  (or
    (for/first ((q-info (in-list p*))
                #:when (equal? (q-info->q-num q-info) key))
      (q-info->rest q-info))
    '()))

(define (prompt-answer* r*)
  (if (null? r*) (raise-argument-error 'prompt-answer "pair?" r*) (last r*)))

(define (prompt-print p* q-key)
  (define rr* (prompt-ref p* q-key))
  (cond
    [(null? rr*)
     (void)]
    [(or
       (ltl->english-part-question? q-key)
       (english->ltl-part-question? q-key))
     (printf "~a~nA. ~a~n" (first rr*) (second rr*))]
    [(traces-part-question? q-key)
     (printf "formula: ~a~n trace: ~a~n answer ~a~n" (first rr*) (second rr*) (third rr*))]
    [else
      (for-each displayln rr*)]))

(define (prompt-print-html p* q-key)
  (define rr* (prompt-ref p* q-key))
  (cond
    [(null? rr*)
     (void)]
    [(or
       (ltl->english-part-question? q-key)
       (english->ltl-part-question? q-key))
     (printf "<table>~n<tr><td>Q</td><td>~a</td></tr>~n<tr><td>A</td><td>~a</td></tr></table>~n" (first rr*) (car (second rr*)))]
    [(traces-part-question? q-key)
     (printf "<table>~n<tr><td>formula:</td><td>~a</td></tr>~n<tr><td>trace:</td><td>~a</td></tr>~n<tr><td>answer</td><td>~a</td></tr></table>~n" (first rr*) (second rr*) (car (third rr*)))]
    [else
      (for-each displayln rr*)]))

(define q-info->q-num car)
(define q-info->rest cdr)

(define (string->seconds str)
  (exact-round (string->number str)))

(define (bytes-count-tab str)
  (for/sum ((c (in-bytes str)))
    (if (eq? c #\tab) 1 0)))

(define (count-tab str)
  (for/sum ((c (in-string str)))
    (if (eq? c #\tab) 1 0)))

(define (parse-tsv tsv)
  (with-input-from-file
    tsv
    (lambda ()
      (for/list ((ln (in-lines)))
        (tab-split ln)))))

(define (tab-split str)
  (string-split str "\t"))

(define (all-distinct? x*)
  (= (length x*)
     (length (remove-duplicates x*))))

(define ((make-row-ref schema) rr kk)
  (or
    (for/first ((r (in-list rr))
                (s (in-list schema))
                #:when (equal? s kk))
      r)
    (raise-arguments-error 'row-ref "key not found in schema" "key" kk "schema" (~e schema) "row" (~e rr))))

(define row->email
  (case-lambda
    ((rr)
     (list-ref rr 6))
    ((rr ref)
     (ref rr "ResponseId"))))

(module+ test
  (test-case "row->email"
    (check-equal? (row->email '("2021-04-09 18:55:52" "2021-04-09 19:55:10" "IP Address" "100" "3557" "anonymous" "917u0cvrthn6oh5a@gmail.com" "R_1rwE5OrOOuIeigR" "EN" "Chrome" "89.0.4389.114" "Macintosh" "1440x900" "5.32" "366.142" "370.408" "12" "If the red light is on then three states from now it must be on " "5 (Highly Confident)" "A possible misconception is the light stays on for all intermediate states" "Two states from now there will be some future state such that the state after it will have the red light on.   To put in more mathematical terms (sorry) but in pseudo English : If the current state is i then there is a state i+2 and a state x such that x is greater than or equal  to i+2 such that in x+1 the red light is on" "3 (Somewhat Confident)" "There could be a lot of confusion about the nesting of the operators and what it means such as thinking that the state after has the red light on or that red light is on 3 states from now." "If the red light will be on now or in some future state that means the blue light must always be on." "4"))
                  "917u0cvrthn6oh5a@gmail.com")
    (check-equal? (row->email '("A" "B" "C")
                              (make-row-ref '("QQ" "ResponseId" "blah")))
                  "B")))

(define filter-ignore-row
  (let* ((black-str "{\"ImportId\":\"startDate\",\"timeZone\":\"America/Denver\"}"))
    (lambda (ds)
      (define row-ref (make-row-ref (dataset-schema ds)))
      (make-dataset
        (dataset-schema ds)
        (dataset-text ds)
        (for/list ((rr (in-list (dataset-row* ds)))
                   #:unless (or (equal? (first rr) black-str)
                                (equal? (row-ref rr "Finished") "False")))
          rr)))))

(define (basic-question? str)
  (regexp-match? #rx"^Q[0-9]+\\.[0-9]+(_[0-9]+)?$" str))

(define (sec->min t)
  (quotient t 60))

(define (correct-traces? student staff)
  (eq? (yes-no->boolean student) staff))

(define (yes-no->boolean str)
  (cond
    [(equal? str "Yes") #true]
    [(equal? str "No") #false]
    [else (raise-argument-error 'yes-no->boolean "(or/c \"Yes\" \"No\")" str)]))

(define (tsv->dataset tsv)
  (let* ((row* (parse-tsv tsv))
         (ds (make-dataset (first row*) (second row*) (drop row* 3)))
         (ds (filter-ignore-row ds)))
    ds))

(define (question-string? str)
  (string-prefix? str "Q"))

(define (ltl->english-part-question? str)
  (string-prefix? str "Q2."))

(define (english->ltl-part-question? str)
  (string-prefix? str "Q4."))

(define (traces-part-question? str)
  (string-prefix? str "Q6."))

(define question-needs-validation*
  '("Q2.23" "Q4.3" "Q4.5" "Q4.7" "Q4.9" "Q4.11" "Q4.13" "Q4.15" "Q4.17" "Q4.19"
    "Q4.21" "Q6.22_1" "Q6.22_2"))

(define (question-needs-validation? str)
  (and (member str question-needs-validation*) #true))

(define question-traces-yes-no*
  '("Q11" "Q13" "Q14" "Q15" "Q16" "Q17" "Q18" "Q19"))

(define (traces-boolean-question? ss)
  (and (member ss question-traces-yes-no*) #true))

(define english->ltl*
  '("Q37" "Q38" "Q39" "Q40" "Q41"))

(define (english->ltl-question? ss)
  (and (member ss english->ltl*) #true))

(define ltl->english*
  '("Q27" "Q28" "Q29" "Q30" "Q31"))

(define (ltl->english-question? ss)
  (and (member ss ltl->english*) #true))

(define extra-miscon*
  '("inexpressible" "syntax error" "precedence" "past-time"))

(define (q+ str n)
  (define-values [pre post]
    (let* ((str* (string-split str ".")))
      (values (car str*) (cadr str*))))
  (format "~a.~a" pre (+ (string->number post) n)))

(define historic-kw* '("once" "'" "historically" "before" "never"))

(define (historic-str? str)
  (for/or ((hh (in-list historic-kw*)))
    (string-contains? str hh)))

;; -----------------------------------------------------------------------------

(module+ main

  #;(check-tabs qualtrics-raw.tsv)
  #;(check-traces qualtrics-raw.tsv)
  #;(check-english->ltl-avg-confidence qualtrics-raw.tsv)
  #;(quick-check-optional (build-path output-dir "per-optional.txt"))

  (print-traces qualtrics-raw.tsv)
  #;(print-ltl->english qualtrics-raw.tsv)
  #;(print-english->ltl qualtrics-raw.tsv)

  (void))

