#lang racket

;; parse.rkt
;;
;; Usage:
;;  racket parse.rkt data.tsv
;;
;; Prints three spreadsheets, each with questions answers and ID info:
;;  1. LE.tsv
;;  2. EL.tsv
;;  3. TB.tsv
;;
;; The sheets are uncoded. Coded versions are online:
;;  https://docs.google.com/spreadsheets/d/15XdyeROZn9rbJDJ8piLNCwYSEqA8-6GsumJGZ14jw-E/edit

(define edit.tsv "edit.tsv")

(define (parse-spreadsheet tsv)
  (call-with-input-file
    tsv
    (lambda (pre-pp)
      (define pp (peeking-input-port pre-pp))
      (define title* (tab-split (string-trim (read-line pp))))
      (define TT (length title*))
      (define (read-row)
        (for/hash ((kk (in-list title*))
                   (idx (in-naturals 1)))
          (define seek-char (if (= idx TT) #\newline #\tab))
          (define vv (read-until pp seek-char))
          (values kk vv)))
      (define prompt* (read-row))
      (void (read-row))
      (let loop ((acc '()))
        (if (eof-object? (peek-char pp))
          (reverse acc)
          (loop (cons (read-row ) acc)))))))

(define (read-until pp chr)
  (with-output-to-string
    (lambda ()
      (let loop ()
        (let* ((nxt (read-char pp)))
          (cond
            [(eof-object? nxt)
             (raise-arguments-error 'read-until "unexpected EOF" "port" pp "seeking" chr)]
            [(eq? nxt chr)
             (void)]
            [else
              (unless (eq? nxt #\return)
                (write-char nxt))
              (loop)]))))))

(define (print-tb row*)
  (define q++* '(
    ("Q11" "Engine" "{GB} {RGB} {RGB} {RGB} {RGB}" "No")
    ("Q45" "next_state ( next_state ( next_state ( Engine ) ) )" "{R} {} {} {R} {}" "Yes")
    ("Q13" "always ( Engine implies next_state ( next_state ( next_state ( Engine ) ) ) )" "{} {RGB} {RGB} {} {RGB}" "Yes")
    ("Q14" "( next_state ( Engine ) ) until ( next_state ( DoorOpen ) )" "{RB} {RB} {RB} {RGB} {B}" "Yes")
    ("Q15" "{ eventually ( Red ) } and { eventually ( Green ) }" "{} {G} {} {} {R}" "Yes")
    ("Q16" "next_state ( next_state ( eventually ( Engine ) ) )" "{RGB} {RGB} {RGB} {RGB} {RGB}" "Yes")
    ("Q17" "( Engine ) until ( Light )" "{R} {R} {R} {R} {R}" "No")
    ("Q18" "eventually ( always ( Engine ) )" "{} {RGB} {} {RGB} {}" "No")
    ("Q19" "always ( Engine implies Light )" "{} {} {} {} {}" "Yes")))
  (define title "Answer	ID	Formula	Trace	Expected")
  (print-sheet "tb.tsv" title row* q++*))

(define (print-el row*)
  (define q++* '(
    ("Q27" "Red implies after ( after ( after ( Red ) ) )" "If Red initially, then Red again in 4th state.")
    ("Q28" "after ( after ( eventually ( after ( Red ) ) ) )" "Red at 4th state or later")
    ("Q29" " { eventually ( Red ) } implies { always ( Blue ) }" "If ever Red, then Blue always from first state")
    ("Q30" " ( ( Red ) until ( Blue ) ) and always ( Red )" "Red forever, Blue eventually")
    ("Q31" "always ( Red implies ( after ( not Red ) and after ( after ( Red ) ) ) )" "If Red then blinks forever")))
  (define title "Answer	ID	Prompt	Expected")
  (print-sheet "el.tsv" title row* q++*))

(define (print-le row*)
  (define q++* '(
    ("Q37" "Whenever the Red light is on, it is off in the next state and on again in the state after that" "G(Red => X(!Red) and XX(Red))")
    ("Q38" "The Red light is on in exactly one state, but not necessarily the first state" "F(Red) and G(Red => X(G(!Red)))")
    ("Q39" "The Red light cannot stay on for three states in a row" "G(Red and X(Red) => XX(!Red))")
    ("Q40" "Whenever the Red light is on, the Blue light will be on then or at some point in the future" "G(Red => F(Blue))")
    ("Q41" "The Red light is on for zero or more states, and then turns off and remains off in the future" "F(!Red) and G(!Red => X(G(!Red)))")))
  (define title "Answer	ID	Prompt	Expected")
  (print-sheet "le.tsv" title row* q++*))

(define (print-sheet fn title row* q++*)
  (with-output-to-file
    fn
    #:exists 'replace
    (lambda ()
      (displayln title)
      (for ((q++ (in-list q++*)))
        (define q (car q++))
        (define other (cdr q++))
        (for ((r (in-list row*)))
          (displayln
            (tab-join
              (list*
                (find-answer r q)
                (find-id r)
                other))))))))

(define (find-answer r q)
  (string-cleanup (row-find r q)))

(define (row-find r k)
  (hash-ref r k))

(define (string-cleanup str)
  (string-replace str "\n" "\\n"))

(define (find-id r)
  (row-find r "ResponseId"))

(define (print-stat row*)
  (printf "STAT~n")
  (printf " num rows: ~a~n" (length row*))
  (printf " final row:~n")
  (pretty-write (last row*))
  (void))

;; ---

(define TAB "\t")

(define (tab-split str)
  (string-split str TAB))

(define (tab-join str*)
  (string-join str* TAB))

;; ---

(module+ main
  (let* ((row* (parse-spreadsheet edit.tsv)))
    #;(print-stat row*)
    (print-tb row*)
    (print-le row*)
    (print-el row*)
    (void)))

